"""
/***************************************************************************

	Author			:Charles B. Cosse 
	
	Email			:ccosse@gmail.com
					
	Website			:www.asymptopia.org
	
	Copyright		:(C) 2002-2007 Asymptopia Software.
	
 ***************************************************************************/
"""


import os, pygame,time
from pygame.locals import *
from random import random

from board import *
from spot import *
from tile import *
from button import *
from player import Player

DEBUG=False

class TMS_Solver:
	
	def __init__(self,game):
		
		self.game=game
		self.tray=None
		self.str2pt=self.game.global_config['scoring']

		self.plets={
			'triplets':[],
			'doublets':[],
			'singlets':[],
			'operator_doublets':self.get_operator_plets(2),
			'operator_singlets':self.get_operator_plets(1),
		}
		
		self.expressions={
		   'triplet_expressions':[],
		   'doublet_expressions':[],
		   'singlet_expressions':[],
		   'wc_triplet_expressions':[],
		   'wc_doublet_expressions':[],
		   'wc_singlet_expressions':[],
		   'wc_wc_doublet_expressions':[],
		}
	
	#This needs modified to handle mult/div:
	def get_operator_plets(self,N):
		
		plets=[]
		available_ops=self.game.global_config['available_operators'][self.game.LEVEL-1]
		
		if(N==1):
			for oidx in range(len(available_ops)):
				plets.append([available_ops[oidx]])
		
		elif(N==2):
			if available_ops.count('-') and available_ops.count('/'):
				plets.append(['-','/'])
				plets.append(['/','-'])
			if available_ops.count('*') and available_ops.count('/'):
				plets.append(['*','/'])
				plets.append(['/','*'])
			if available_ops.count('+') and available_ops.count('/'):
				plets.append(['+','/'])
				plets.append(['/','+'])
			if available_ops.count('+') and available_ops.count('*'):
				plets.append(['+','*'])
				plets.append(['*','+'])
			if available_ops.count('-') and available_ops.count('*'):
				plets.append(['-','*'])
				plets.append(['*','-'])
			if available_ops.count('+') and available_ops.count('-'):
				plets.append(['+','-'])
				plets.append(['-','+'])
			if available_ops.count('/'):
				plets.append(['/','/'])
			if available_ops.count('*'):
				plets.append(['*','*'])
			if available_ops.count('-'):
				plets.append(['-','-'])
			if available_ops.count('+'):
				plets.append(['+','+'])

		return(plets)
	
	def getStringValues(self):
		str_values=[]
		spots=self.tray.get_spots()#not sorted 1-10,so sort:
		nnumbers=self.game.NNUMBERS
		newspots=[]
		for spotidx in range(nnumbers):
			newspots.append(0)#=[0,0,0,0,0,0]
		
		ntrayspots=self.game.NTRAYSPOTS
		while len(spots)>ntrayspots-nnumbers:
			for spot in spots:
				if spot.getMN()[1]<nnumbers:
					newspots[spot.getMN()[1]]=spot
					spots.remove(spot)
				#else:print spot.getMN()
		
		for dummy in range(len(newspots)):
			str_values.append(newspots[dummy].guest.str_val)
		return(str_values)
	
	def getAllStringValues(self):
		str_values=[]
		spots=self.tray.get_spots()#not sorted 1-10,so sort:
		ntrayspots=self.game.NTRAYSPOTS
		newspots=[]
		for spotidx in range(ntrayspots):
			newspots.append(0)#=[0,0,0,0,0,0,0,0,0,0]
		while len(spots)>0:
			for spot in spots:
				newspots[spot.getMN()[1]]=spot
				spots.remove(spot)
		for dummy in range(len(newspots)):
			str_values.append(newspots[dummy].guest.str_val)
		return(str_values)

	def cycle_vals(self,vals):
		tmp=vals.pop()
		vals.insert(int(random()*len(vals)),tmp)
		return(vals)
	
	def get3x2x1x(self,N):
		#N=1,2,3 ~ singlets,doublets,tripplets
		#all unique index-triplets in set of 6 Tiles:
		#if 2 "5"'s in tiles, two "5"-singlets returned, etc..
		plets=[]
		num_times_cycled=0
		str_vals=self.getStringValues()
		#print str_vals
		NumNumbers=6
		while(num_times_cycled<1000):#print ratio when added vs idx on this before v2.0 release
			for idx in range(0,NumNumbers-N):
				i_plet=[]
				s_plet=[]
				for jdx in range(N):
					i_plet.append(float(str_vals[idx+jdx]))#changed to "float" v2.0
				i_plet.sort()
				for jdx in range(N):
					s_plet.append(`i_plet[jdx]`)
				
				if plets.count(s_plet)==0:plets.append(s_plet)
				elif(N==1 and plets.count(s_plet)<str_vals.count(s_plet[0])):
					plets.append(s_plet)
				
			str_vals=self.cycle_vals(str_vals)
			num_times_cycled=num_times_cycled+1
		return(plets)
	
	def evaluate(self,expr):	
		#print expr
		str=''
		for idx in range(len(expr)):
			str=str+expr[idx]
		
		try:
			val=eval(str)
			#print str,val
		except:return(None)
		return(val)

	def get3xPermutations(self,tripplet):#receives single triplet list of len=3
		#print 'tripplet:',tripplet
		p=[[1,2,3],[1,3,2],[2,1,3],[2,3,1],[3,1,2],[3,2,1]]
		px3=[]#list of permuted "t"riplets
		for pidx in range(len(p)):
			tx3=[]#a permuted "t"riplet
			for tidx in range(3):
				#print 'pidx:',pidx,'  tidx:',tidx,'  p[pidx][tidx]-1=',p[pidx][tidx]-1,'  tripplet[p[pidx][tidx]-1]=',tripplet[p[pidx][tidx]-1]
				tx3.append(tripplet[p[pidx][tidx]-1])
			if px3.count(tx3)==0:px3.append(tx3)
			#print 'px3:',px3
		return(px3)	
		
	def get2xPermutations(self,doublet):
		p=[[1,2],[2,1]]
		px2=[]#list of permuted doublets
		for pidx in range(len(p)):
			dx2=[]#permuted doublet
			for tidx in range(2):
				dx2.append(doublet[p[pidx][tidx]-1])
			if px2.count(dx2)==0:px2.append(dx2)
		return(px2)	
			
	def generate_expressions(self):
		
		self.tray=self.game.players[self.game.player_idx].tray
		
		while len(self.plets['triplets']):self.plets['triplets'].pop()
		while len(self.plets['doublets']):self.plets['doublets'].pop()
		while len(self.plets['singlets']):self.plets['singlets'].pop()
		
		if self.game.global_config['available_plets'].count('triplets'):self.plets['triplets']=self.get3x2x1x(3)
		if self.game.global_config['available_plets'].count('doublets'):self.plets['doublets']=self.get3x2x1x(2)
		if self.game.global_config['available_plets'].count('singlets'):self.plets['singlets']=self.get3x2x1x(1)
		
		while len(self.expressions['triplet_expressions']):x=self.expressions['triplet_expressions'].pop();del x
		while len(self.expressions['doublet_expressions']):x=self.expressions['doublet_expressions'].pop();del x
		while len(self.expressions['singlet_expressions']):x=self.expressions['singlet_expressions'].pop();del x
		while len(self.expressions['wc_triplet_expressions']):x=self.expressions['wc_triplet_expressions'].pop();del x
		while len(self.expressions['wc_doublet_expressions']):x=self.expressions['wc_doublet_expressions'].pop();del x
		while len(self.expressions['wc_singlet_expressions']):x=self.expressions['wc_singlet_expressions'].pop();del x
		while len(self.expressions['wc_wc_doublet_expressions']):x=self.expressions['wc_wc_doublet_expressions'].pop();del x
		
		max_replacements=self.game.global_config['max_replacements'][self.game.LEVEL-1]
		
		if max_replacements>=0 and self.game.global_config['available_plets'].count('triplets'):
			for pdx in range(len(self.plets['triplets'])):
				permutations=self.get3xPermutations(self.plets['triplets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_doublets'])):
						expr=[perm[0],self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],perm[2]]#save these
						value=self.evaluate(expr)
						if not value:continue
						self.expressions['triplet_expressions'].append([pdx,odx,expr,value])
			

		if max_replacements>=1 and self.game.global_config['allow_wc3x3']:
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
			#"""deactivating: this works, but takes too long (testing, not generating (here)) by brute force.
			
			#wc_idx=0
			for pdx in range(0,len(self.plets['triplets'])):
				permutations=self.get3xPermutations(self.plets['triplets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_doublets'])):
						for val in range(16):
							expr=[`val`,self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],perm[2]]
							value=self.evaluate(expr)
							if not value:continue
							expr=['WC:'+`val`+'.0',self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],perm[2]]
							self.expressions['wc_triplet_expressions'].append([pdx,odx,expr,value])
			#wc_idx=1
			for pdx in range(0,len(self.plets['triplets'])):
				permutations=self.get3xPermutations(self.plets['triplets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_doublets'])):
						for val in range(16):
							expr=[perm[0],self.plets['operator_doublets'][odx][0],`val`,self.plets['operator_doublets'][odx][1],perm[2]]
							value=self.evaluate(expr)
							if not value:continue
							expr=[perm[0],self.plets['operator_doublets'][odx][0],'WC:'+`val`+'.0',self.plets['operator_doublets'][odx][1],perm[2]]
							self.expressions['wc_triplet_expressions'].append([pdx,odx,expr,value])
			#wc_idx=2
			for pdx in range(0,len(self.plets['triplets'])):
				permutations=self.get3xPermutations(self.plets['triplets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_doublets'])):
						for val in range(16):
							expr=[perm[0],self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],`val`]
							value=self.evaluate(expr)
							if not value:continue
							expr=[perm[0],self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],'WC:'+`val`+'.0']
							self.expressions['wc_triplet_expressions'].append([pdx,odx,expr,value])
			#wc_idx~operator[0]
			for pdx in range(0,len(self.plets['triplets'])):
				permutations=self.get3xPermutations(self.plets['triplets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_doublets'])):
						expr=[perm[0],self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],perm[2]]
						value=self.evaluate(expr)
						if not value:continue
						expr=[perm[0],'WC:'+self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],perm[2]]
						self.expressions['wc_triplet_expressions'].append([pdx,odx,expr,value])
			#wc_idx~operator[1]
			for pdx in range(0,len(self.plets['triplets'])):
				permutations=self.get3xPermutations(self.plets['triplets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_doublets'])):
						expr=[perm[0],self.plets['operator_doublets'][odx][0],perm[1],self.plets['operator_doublets'][odx][1],perm[2]]
						value=self.evaluate(expr)
						if not value:continue
						expr=[perm[0],self.plets['operator_doublets'][odx][0],perm[1],'WC:'+self.plets['operator_doublets'][odx][1],perm[2]]
						self.expressions['wc_triplet_expressions'].append([pdx,odx,expr,value])
			
			#print 'finished wc_triplet_expressions generation:',len(triplet_expressions),len(wc_triplet_expressions)
			#"""
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		if max_replacements>=0:
			for pdx in range(0,len(self.plets['doublets'])):
				permutations=self.get2xPermutations(self.plets['doublets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_singlets'])):
						expr=[perm[0],self.plets['operator_singlets'][odx][0],perm[1]]
						value=self.evaluate(expr)
						if not value:continue
						self.expressions['doublet_expressions'].append([pdx,odx,expr,value])
			
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		if max_replacements>=1:
			#wc_idx=0
			for pdx in range(0,len(self.plets['singlets'])):
				for odx in range(len(self.plets['operator_singlets'])):
					for val in range(len(self.game.global_config['available_numbers'][self.game.LEVEL-1])):
						expr=[`val`+'.0',self.plets['operator_singlets'][odx][0],self.plets['singlets'][pdx][0]]
						value=self.evaluate(expr)
						if not value:continue
						expr=['WC:'+`val`+'.0',self.plets['operator_singlets'][odx][0],self.plets['singlets'][pdx][0]]
						self.expressions['wc_doublet_expressions'].append([pdx,odx,expr,value])

			
			#wc_idx=1
			for pdx in range(0,len(self.plets['singlets'])):
				for odx in range(len(self.plets['operator_singlets'])):
					for val in range(len(self.game.global_config['available_numbers'][self.game.LEVEL-1])):
						expr=[self.plets['singlets'][pdx][0],self.plets['operator_singlets'][odx][0],`val`+'.0']
						value=self.evaluate(expr)
						if not value:continue
						expr=[self.plets['singlets'][pdx][0],self.plets['operator_singlets'][odx][0],'WC:'+`val`+'.0']
						self.expressions['wc_doublet_expressions'].append([pdx,odx,expr,value])
			
			
			#wc~operator:
			for pdx in range(0,len(self.plets['doublets'])):
				permutations=self.get2xPermutations(self.plets['doublets'][pdx])#permutations of idx+1 (i.e. 1,2,3 rather than 0,1,2)
				for perm in permutations:
					for odx in range(len(self.plets['operator_singlets'])):
						expr=[perm[0],self.plets['operator_singlets'][odx][0],perm[1]]
						value=self.evaluate(expr)
						if not value:continue
						expr=[perm[0],'WC:'+self.plets['operator_singlets'][odx][0],perm[1]]
						self.expressions['wc_doublet_expressions'].append([pdx,odx,expr,value])
			
			
			#SHUFFLE EXPRESSIONS TO GET EVEN REPRESENTATION
			for dummy in range(len(self.expressions['wc_doublet_expressions'])):
				insert_idx=int( (  len(self.expressions['wc_doublet_expressions'])  -2  ) *random() )
				expr2insert=self.expressions['wc_doublet_expressions'].pop()
				self.expressions['wc_doublet_expressions'].insert(insert_idx,expr2insert)
			
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		if max_replacements>=2:
			#wc_idx=1 AND wc_idx=2
			for v1 in range(len(self.game.global_config['available_numbers'][self.game.LEVEL-1])):
				for v2 in range(len(self.game.global_config['available_numbers'][self.game.LEVEL-1])):
					
					for odx in range(len(self.plets['operator_singlets'])):
						
						expr=[`v1`+'.0',self.plets['operator_singlets'][odx][0],`v2`+'.0']
						value=self.evaluate(expr)
						if not value:continue
						expr=['WC:'+`v1`+'.0',self.plets['operator_singlets'][odx][0],'WC:'+`v2`+'.0']
						self.expressions['wc_wc_doublet_expressions'].append([0,odx,expr,value])#what did "pdx"(now "0") do?
			
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		if max_replacements>=0:
			for pdx in range(len(self.plets['singlets'])):
				expr=[self.plets['singlets'][pdx][0]]
				value=self.evaluate(expr[0])
				if not value:continue
				self.expressions['singlet_expressions'].append([pdx,None,expr,value])
			
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		if max_replacements>=1:
			#wc_idx=0
			for pdx in range(len(self.game.global_config['available_numbers'][self.game.LEVEL-1])):
				expr=['WC:'+`pdx`+'.0']
				value=pdx
				if not value:continue
				self.expressions['wc_singlet_expressions'].append([pdx,None,expr,value])
			
			#+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
		
		self.game.handle_mouse()
		if self.game.STOP_RUNNING!=0:return None
		

	def construct_submission(self,lhs_expressions,rhs_expressions):
		
		for key in self.expressions.keys():
			if lhs_expressions==self.expressions[key]:lhs=key
			if rhs_expressions==self.expressions[key]:rhs=key
			
		if DEBUG:print "%s(%d) * %s(%d) = %e"%(lhs,len(lhs_expressions),rhs,len(rhs_expressions),len(lhs_expressions)*len(rhs_expressions))
		
		stringValues=self.getAllStringValues()
		str2pt=self.str2pt
		score=0
		self.game.handle_mouse()
		if self.game.STOP_RUNNING!=0:return None
		
		self.game.queue_thinking_maneuver()
		self.game.update()
		self.game.update()
		self.game.update()
		
		for idx1 in range(len(lhs_expressions)):
			for idx2 in range(len(rhs_expressions)):
				if lhs_expressions[idx1][3]==rhs_expressions[idx2][3]:#check equality of expressions before equating
					#print 'combining:',lhs_expressions[idx1],rhs_expressions[idx2]
					combinedList=[]
					for idx in range(len(lhs_expressions[idx1][2])):
						#print 'idx1=',idx1,' idx=',idx
						combinedList.append(lhs_expressions[idx1][2][idx])
					combinedList.append('=')
					for idx in range(len(rhs_expressions[idx2][2])):
						combinedList.append(rhs_expressions[idx2][2][idx])
					
					ok=1
					for elem in combinedList:
						count=combinedList.count(elem)
						if elem.count('WC:'):#Then we don't want to check for it in stringValues!!
							pass
						elif stringValues.count(elem)<count:
							#print combinedList,elem,count,stringValues.count(elem)
							ok=0
					
					if ok==1:
						rlist=self.game.localizer.localize(combinedList)
						if rlist:
							score=0
							for cidx in range(len(combinedList)):
								if combinedList[cidx][:3]=='WC:':
									if self.game.global_config['points_for_old']:score+=str2pt[combinedList[cidx][3:]]
									self.game.num_replacements+=1
								else:score+=str2pt[combinedList[cidx]]
								
							pidx=self.game.player_idx
							self.game.players[pidx].score+=score
							self.game.last_points=score
							return(rlist)
							
					else:del combinedList
					
		self.game.last_points=score
		return(None)
							
	def generate_options(self):
	
		triplet_expressions			=self.expressions['triplet_expressions']
		doublet_expressions			=self.expressions['doublet_expressions']
		singlet_expressions			=self.expressions['singlet_expressions']
		wc_triplet_expressions		=self.expressions['wc_triplet_expressions']
		wc_doublet_expressions		=self.expressions['wc_doublet_expressions']
		wc_wc_doublet_expressions	=self.expressions['wc_wc_doublet_expressions']
		wc_singlet_expressions		=self.expressions['wc_singlet_expressions']
		operator_doublets			=self.plets['operator_doublets']
		operator_singlets			=self.plets['operator_singlets']
		
		LEVEL=self.game.LEVEL
		#print 'tux generating options'	

		self.options=[]#the result of this function
		tray=self.tray
		spots=self.tray.get_spots()
		
		#board get 1 copy 1x str_val array in preparation of numerous brute-force localization attempts:
		self.game.localizer.update_board_map()
		
		num_commited=self.game.board.get_num_commited()
		
		max_replacements=self.game.global_config['max_replacements'][self.game.LEVEL-1]
		allow_wc3x3=self.game.global_config['allow_wc3x3']
		t0=time.time()
		
		rlist=None
		
		if num_commited>=7 and max_replacements>=4:
			rlist=self.construct_submission(wc_wc_doublet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
		if num_commited>=7 and max_replacements>=3:
			if allow_wc3x3:rlist=self.construct_submission(wc_triplet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			if allow_wc3x3:rlist=self.construct_submission(wc_wc_doublet_expressions,wc_triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(wc_doublet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_wc_doublet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(wc_wc_doublet_expressions,wc_singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_singlet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			if DEBUG:print 'done max_replacements=3'
		
		if num_commited>=4 and max_replacements>=2:
			rlist=self.construct_submission(wc_triplet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_doublet_expressions,wc_triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(triplet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_wc_doublet_expressions,triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(wc_triplet_expressions,wc_singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_singlet_expressions,wc_triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(wc_doublet_expressions,wc_singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_singlet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0

			rlist=self.construct_submission(doublet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_wc_doublet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0

			rlist=self.construct_submission(wc_wc_doublet_expressions,singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(singlet_expressions,wc_wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0

			rlist=self.construct_submission(wc_doublet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			if DEBUG:print 'done max_replacements=2'
		
		
		if num_commited>=1 and max_replacements>=1:
			rlist=self.construct_submission(triplet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_doublet_expressions,triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(wc_triplet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(doublet_expressions,wc_triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(doublet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_doublet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(doublet_expressions,wc_singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_singlet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(singlet_expressions,wc_doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(wc_doublet_expressions,singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(wc_singlet_expressions,singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(singlet_expressions,wc_singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			if DEBUG:print 'done max_replacements=1'
			

		if max_replacements>=0 and (self.game.global_config['allow_disconnected'] or num_commited==0):
			rlist=self.construct_submission(triplet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(doublet_expressions,triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(triplet_expressions,singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(singlet_expressions,triplet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(doublet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(singlet_expressions,doublet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			rlist=self.construct_submission(doublet_expressions,singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
			rlist=self.construct_submission(singlet_expressions,singlet_expressions)
			if rlist:return(rlist)
			if DEBUG:print time.time()-t0
			
		return None
