"""
/***************************************************************************

	Author 			:Charles B. Cosse 
	
	Email			:ccosse@asymptopia.org
	
	Website			:http://www.asymptopia.org
	
	Copyright		:(C) 2004-2006 Asymptopia Software.
	
 ***************************************************************************/
"""
import sys,os,string,time
from random import random
from string import *

import pygame
from pygame.locals import *

from eduapp import *
from player import *
from actor import *
from board import *
from spot import *
from button import *
from recorder import *
from projectile import *

from tms_solver import *
from tms_validator import *
from tms_localizer import *

if not pygame.font: print 'Warning, fonts disabled'
if not pygame.mixer: print 'Warning, sound disabled'

DEBUG=False

class TMS_Window(EduAppWindow):
	
	def __init__(self,parent,username,MODE):
		
		EduAppWindow.__init__(self,parent,username,MODE)
		
		self.parent=parent
		self.MODE=MODE
		self.players=None
		self.player_idx=None
		self.actors=None
		self.player_idx=None
		self.tuxturn=None
		self.localizer=None
		self.validator=None
		self.screen=None
		self.board=None
		self.background=None
		self.buttons=None
		self.loginbuttons=None
		self.boardspots=None
		self.tray_spots=None
		self.cannot_move_count=None
		self.trayspots=None
		self.submission=None
		self.submissionspots=None
		self.target=None
		self.animation_in_progress=None
		self.AnimatedTiles=None
		
		self.l1_button=None
		self.l2_button=None
		self.l3_button=None
		self.l4_button=None
		
		self.login_button=None
		self.logout_button=None
		self.credit_button=None
		self.skip_button=None
		self.ready_button=None
		
		self.level_indicator=None
		self.pyfont=None
		self.NTRAYSPOTS=None
		self.NNUMBERS=None
		self.LEVEL=None
		self.high_score_list=None
		self.scoring=None
		self.STOP_RUNNING=0
		self.screenshot_counter=0
		self.pickup_sounds	=None
		self.release_sounds	=None
		self.lockin_sounds	=None
		self.bounce_sounds	=None
		self.lose_sounds	=None
		self.win_sounds		=None

		self.recorder=None
		self.t0=None
		self.t_session=None
		self.t_last=None
		self.dt_last=None
		self.last_points=None
		self.solver_submission=None
		self.num_new_tiles=None
		self.num_replacements=None
		self.correct=None
		self.move_number=None

		self.fwoverlay=None
		
	def fw_show(self):
		
		self.fwoverlay=FWOverlay(
			self.screen,
			self.global_config['fw_dt'],
			self.global_config['fw_dt_launch'],
			self.global_config['fw_scaleFactor']
		)
		
		self.STOP_RUNNING=0
		
		self.fwoverlay.make_volley(self.global_config['fw_volley_size'])
		
		msg01=self.congrats_font.render(
			"TuxMathScrabble",1,
			self.global_config['congrats_fg'],(0,0,0)
		)
		msg01.set_colorkey((0,0,0))
		fw01=msg01.get_rect()[2]
		fh01=msg01.get_rect()[3]
		xp01=(self.W-fw01)/2
		yp01=(self.H)/2-fh01
		
		msg02=self.pyfont.render(
			"Asymptopia Software",1,
			self.global_config['congrats_fg'],(0,0,0)
		)
		msg02.set_colorkey((0,0,0))
		fw02=msg02.get_rect()[2]
		fh02=msg02.get_rect()[3]
		xp02=(self.W-fw02)/2
		yp02=(self.H)/2
		
		self.pyfont.set_bold(1)
		msg03=self.pyfont.render(
			"Software@theLimit",1,
			self.global_config['congrats_fg'],(0,0,0)
		)
		msg03.set_colorkey((0,0,0))
		fw03=msg03.get_rect()[2]
		fh03=msg03.get_rect()[3]
		xp03=(self.W-fw03)/2
		yp03=(self.H)/2+fh02
		
		
		count=0
		while len(self.fwoverlay.projectiles):
			
			if self.STOP_RUNNING!=0:
				self.onexit()
				return(0,self.STOP_RUNNING)
			
			self.handle_mouse()
			for event in pygame.event.get(KEYDOWN):
				if event.key == K_ESCAPE:
					pygame.quit()
					sys.exit()
				elif event.key == K_F4:
					pygame.display.toggle_fullscreen()
					self.amfullscreen*=-1
					
			pygame.event.clear()
			
			self.screen.fill((0,0,0))
			self.fwoverlay.tick()

			self.buttons.draw(self.screen)
			self.loginbuttons.draw(self.screen)
			self.screen.blit(msg01,(xp01,yp01))
			self.screen.blit(msg02,(xp02,yp02))
			self.screen.blit(msg03,(xp03,yp03))
			
			pygame.display.flip()
			
			"""
			oufname="%03d.bmp"%(count)
			pygame.image.save(pygame.display.get_surface(),oufname)
			count+=1
			"""
			
		return(0,-2)
		
		
	def onexit(self):
		if self.MODE==1:
			self.user_config['high_score']+=self.players[1].score
			unused_highscore=self.save()

	def draw_tiles(self):
		
		available_numbers=self.global_config['available_numbers']
		available_operators=self.global_config['available_operators']
		spots=self.players[self.player_idx].tray.get_spots()
		
		for spot in spots:
			
			if spot.occupied():continue
			
			if (spot.getMN()[1]<self.NNUMBERS):
				int_val=int(len(available_numbers[self.LEVEL-1])*random())
				str_val=`int_val`+'.0'
				
				if self.player_idx==0:numname="%s%s.%s"%(self.global_config['tux_tile_prefix'],available_numbers[self.LEVEL-1][int_val],'gif')
				else:numname="%s.%s"%(available_numbers[self.LEVEL-1][int_val],'gif')
				
				if self.player_idx==0:fname=os.path.join(self.global_config['imgdir'],numname)
				else:fname=os.path.join(self.global_config['imgdir'],numname)
			
			elif spot.getMN()[1]<self.NTRAYSPOTS-1:
				
				idx=int(len(available_operators[self.LEVEL-1])*random())#add,subtract,mult,divide
				int_val=None
				
				opname=None
				if   available_operators[self.LEVEL-1][idx]=='+':opname='new_plus';str_val='+'
				elif available_operators[self.LEVEL-1][idx]=='-':opname='new_minus';str_val='-'
				elif available_operators[self.LEVEL-1][idx]=='*':opname='new_mult';str_val='*'
				elif available_operators[self.LEVEL-1][idx]=='/':opname='new_divide';str_val='/'
				
				if self.player_idx==0:opname="tux_%s.%s"%(opname,'gif')
				else: opname="%s.%s"%(opname,'gif')
				fname=os.path.join(self.global_config['imgdir'],opname)
			
			else:
				if self.player_idx==0:fname=self.global_config['tux_equal_imgpath']
				else:fname=self.global_config['equal_imgpath']
				int_val=None
				str_val='='
			
			#ptval=self.user_config['str2pt'][str_val]	
			ptval=self.scoring[str_val]
			
			tile=Tile(fname,int_val,str_val,ptval)
			spot.take_guest(tile,1)
			tile.saved_center=spot.rect.center
			

	def setup_common(self):	
		
		self.STOP_RUNNING=0
		pygame.event.set_blocked(MOUSEMOTION)#Wow! this helps!
		
		sitepkgdir=self.parent.env.sitepkgdir

		self.scoring=self.global_config['scoring']
		
		#Make PyGame Window...
		screen=None
		if self.amfullscreen<0:screen = pygame.display.set_mode((self.W,self.H))
		else:screen=pygame.display.set_mode((self.W,self.H),pygame.FULLSCREEN)
		
		pygame.display.set_caption(self.global_config['title_string'])
		pygame.mouse.set_visible(1)
		

		#Try to load sounds:
		try:
			self.pickup_sounds	=[pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','pickup01.wav')),]
			self.release_sounds	=[pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','release01.wav')),]
			self.lockin_sounds	=[pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','lockin01.wav')),]
			self.bounce_sounds	=[pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','bounce01.wav')),]
			self.lose_sounds	=[pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','lose01.wav')),]
			self.win_sounds		=[
									pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','win01.wav')),
									pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','win02.wav')),
									pygame.mixer.Sound(os.path.join(sitepkgdir,self.global_config['appdir'],'Sounds','win03.wav')),
						 		 ]
		except:
			self.global_config['soundon']=0


		####Buttons
		button_imgpaths={
				'login_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','login_button.gif'),
				'logout_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','logout_button.gif'),
				'credit_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','credits_button.gif'),
				'skip_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','skip_button.gif'),
				'ready_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','ready_button.gif'),
				'l1_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','level_one.gif'),
				'l2_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','level_two.gif'),
				'l3_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','level_three.gif'),
				'l4_button':os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','level_four.gif'),
				}
		self.login_button=Button(button_imgpaths['login_button'])
		self.login_button.rect.center=(self.global_config['WIN_W']-70,self.global_config['WIN_H']-285)
		self.logout_button=Button(button_imgpaths['logout_button'])
		self.logout_button.rect.center=(self.global_config['WIN_W']-70,self.global_config['WIN_H']-285)
		self.credit_button=Button(button_imgpaths['credit_button'])
		self.credit_button.rect.center=(self.global_config['WIN_W']-70,self.global_config['WIN_H']-235)
		self.skip_button=Button(button_imgpaths['skip_button'])
		self.skip_button.rect.center=(self.global_config['WIN_W']-70,self.global_config['WIN_H']-135)
		self.ready_button=Button(button_imgpaths['ready_button'])
		self.ready_button.rect.center=(self.global_config['WIN_W']-70,self.global_config['WIN_H']-185)
		
		self.l1_button=Button(button_imgpaths['l1_button'])
		self.l1_button.rect.center=(65,self.global_config['WIN_H']-135)
		self.l2_button=Button(button_imgpaths['l2_button'])
		self.l2_button.rect.center=(65,self.global_config['WIN_H']-185)
		self.l3_button=Button(button_imgpaths['l3_button'])
		self.l3_button.rect.center=(65,self.global_config['WIN_H']-235)
		self.l4_button=Button(button_imgpaths['l4_button'])
		self.l4_button.rect.center=(65,self.global_config['WIN_H']-285)
		
		buttongroup=pygame.sprite.Group([self.logout_button,self.ready_button,self.skip_button,self.credit_button,self.l1_button,self.l2_button,self.l3_button,self.l4_button])
		buttons=pygame.sprite.RenderPlain(buttongroup)
		loginbuttongroup=pygame.sprite.Group([self.login_button,])
		loginbuttons=pygame.sprite.RenderPlain(loginbuttongroup)


				
		#Load Fonts
		pyfont=pygame.font.Font(os.path.join(sitepkgdir,self.global_config['appdir'],'Font','freesansbold.ttf'),12)
		self.congrats_font=pygame.font.Font(self.global_config['congrats_font'],self.global_config['congrats_fontsize'])
		self.splash_font=pygame.font.Font(self.global_config['congrats_font'],self.global_config['congrats_fontsize'])
			
		self.screen=screen
		self.buttons=buttons
		self.loginbuttons=loginbuttons
		self.loginbuttons=loginbuttons
		self.pyfont=pyfont
		self.target=pygame.sprite.RenderClear()#drag-n-drop object	
		

		

	def run(self,LEVEL,high_score_list):
		self.LEVEL=LEVEL
		self.high_score_list=high_score_list
		
		#Make the GameBoard...
		board=Board(
			self.global_config['M'],
			self.global_config['N'],
			self.global_config['WIN_W']/2+10,
			self.global_config['WIN_H']/2,
			None,
			os.path.join(self.parent.env.sitepkgdir,self.global_config['appdir'],'Images','new_spot.gif')
		)
		
		boardspots=pygame.sprite.RenderPlain(board.get_spots())

		
		if self.LEVEL>=3:
			NTRAYSPOTS=12
			NNUMBERS=7
		else:
			NTRAYSPOTS=10
			NNUMBERS=6

		self.NTRAYSPOTS=NTRAYSPOTS
		self.NNUMBERS=NNUMBERS

		background = pygame.Surface(self.screen.get_size())
		background = background.convert()
		background.fill(self.global_config['bgcolor'])#green

		level_indicator = pygame.Surface((10,10))
		level_indicator = level_indicator.convert()
		level_indicator.fill(self.global_config['indicator_fill'])

		self.submission=Board(self.global_config['M'],self.global_config['N'],self.global_config['WIN_W']/2+10,self.global_config['WIN_H']/2,None,os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','new_spot.gif'))
		self.submissionspots=pygame.sprite.RenderPlain()


		validator=TMS_Validator(board,self)#self passed for score update
		self.validator=validator
		localizer=TMS_Localizer(board,self)#self passed for score update
		self.localizer=localizer
		
		trays=[]
		trays.append(	Board(1,NTRAYSPOTS,self.global_config['WIN_W']/2,45,None,os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','new_spot.gif'))	)
		trays.append(	Board(1,NTRAYSPOTS,self.global_config['WIN_W']/2,self.global_config['WIN_H']-40,None,os.path.join(self.sitepkgdir,self.global_config['appdir'],'Images','new_spot.gif'))	    )
		
		solver=TMS_Solver(self)
		
		
		self.player_idx=0
		self.players=[]#players have name,actor and solver
		self.tray_spots=[]
		player_names=self.global_config['player_names']
		player_homes=self.global_config['player_homes']
		for pidx in range(len(player_names)):
			indir=os.path.join(self.parent.env.sitepkgdir,self.global_config['appdir'],'Actors',player_names[pidx])
			actr=Actor(indir,self.global_config['tsleep_animation'],self.W,self.H,player_homes[pidx][0],player_homes[pidx][1],player_homes[pidx][2],None)
			self.players.append(Player(player_names[pidx],trays[pidx],actr,solver))
			self.tray_spots.append(pygame.sprite.RenderPlain(self.players[pidx].tray.get_spots()))
			
		self.players[0].actor.queue('welcome')
		self.players[1].actor.queue('welcome')
		if self.MODE==1:self.players[1].actor.queue('get_up')
		
		
		
		self.AnimatedTiles=[]
		tiledir=os.path.join(self.parent.env.sitepkgdir,self.global_config['appdir'],'Tiles')
		for m in range(self.global_config['M']):
			row=[]
			for n in range(self.global_config['N']):
				xc=board.get_spotMN(m,n).rect.center[0]
				yc=board.get_spotMN(m,n).rect.center[1]
				z=0
				tile=Actor(tiledir,self.global_config['tsleep_animation'],self.W,self.H,xc,yc,z,None)
				row.append(tile)
				#tile.queue('hula')
			self.AnimatedTiles.append(row)
			
		
		self.user_config['IDNsession']+=1
		self.high_score_list=high_score_list

		background = pygame.Surface(self.screen.get_size())
		background = background.convert()
		background.fill(self.global_config['bgcolor'])#green

		self.tuxturn=0
		self.animation_in_progress=0
		self.localizer=localizer
		self.validator=validator
		self.board=board
		self.background=background
		self.boardspots=boardspots
		
		for self.player_idx in range(2):self.draw_tiles()
		self.cannot_move_count=0
		
		self.recorder=EduAppRecorder(self,self.global_config,self.user_config)
		self.players[0].recording=self.global_config['record_player_0']
		self.players[1].recording=self.global_config['record_player_1']
		self.t0=time.time()
		self.t_session=0
		self.t_last=self.t0
		self.dt_last=0
		self.move_number=0
		self.level_indicator=level_indicator
		
		self.STOP_RUNNING=0
		


		while 1:
			
			if self.cannot_move_count>=2:
				self.STOP_RUNNING=-2
				if self.global_config['record_screenshots']:self.take_screenshot()
				
			if self.STOP_RUNNING!=0:
				self.onexit()
				return(self.players[1].score,self.STOP_RUNNING)
			
			self.update()
			
			for self.player_idx in range(2):
				
				if self.player_idx==1 and self.MODE!=0:continue
				elif self.player_idx==0 and not self.tuxturn:continue
				
				
				while self.animation_in_progress:
				   self.update()
				   if self.STOP_RUNNING!=0:
				   	self.onexit()
					return(self.players[0].score,self.STOP_RUNNING)
				
				self.players[self.player_idx].actor.queue('get_up');self.update()
				
				while self.animation_in_progress:
					self.update()
					if self.STOP_RUNNING!=0:
						self.onexit()
						return(self.players[0].score,self.STOP_RUNNING)
				
				
				self.players[self.player_idx].solver.generate_expressions()
				solver_finished=0
			
				while solver_finished!=1:#might not need to be loop now
					self.num_replacements=0
					self.solver_submission=self.players[self.player_idx].solver.generate_options()
					
					rlist=self.solver_submission
					if DEBUG:print rlist
					
					if self.solver_submission==None:
						solver_finished=1
						self.players[self.player_idx].actor.queue('go_back');self.update()
						self.cannot_move_count+=1

						self.move_number+=1
						self.recorder_checkin()
						
						
						if self.player_idx==1:self.tuxturn=1
						else:self.tuxturn=0
						break
					
					else:
						if rlist[0][0]==rlist[1][0]:exprtype='row'
						else:exprtype='col'
						try:
							for idx in range(len(rlist)):
								
								tile=self.players[self.player_idx].tray.get_guest_by_str(rlist[idx][0])
								spot=self.submission.take_guestMN(tile,rlist[idx][1],rlist[idx][2])
								self.submissionspots.add(spot)	
								
						except:print 'error line 388'
							
						#commit submission
						board.increment_num_commited()
						self.cannot_move_count=0
						
						self.players[self.player_idx].actor.queue('celebrate');
						if self.player_idx==0:self.players[1].actor.queue('get_tougher');
						else:self.players[0].actor.queue('get_tougher');
															
						for spot in self.submissionspots.sprites():
							for board_spot in self.board.get_spots():
								if board_spot.rect.center==spot.rect.center:# M,N-Matching
									tile=spot.pop_guest()
									board_spot.take_guest(tile,1)
									board_spot.lock()
									self.submission.remove(spot)
									
									#REACTION TO WIN MANEUVERS:
									self.AnimatedTiles[board_spot.getMN()[0]][board_spot.getMN()[1]].queue('celebrate')
									self.update()
									
						self.move_number+=1
						self.recorder_checkin()
								
						self.submissionspots.empty()
						self.draw_tiles()
							
						self.players[self.player_idx].actor.queue('go_back');self.update()
						solver_finished=1
						if self.player_idx==1:self.tuxturn=1
						else:
							self.tuxturn=0
							if self.MODE!=0:self.players[1].actor.queue('get_up')

	def handle_mouse(self):	

		for event in pygame.event.get():
			
			if event.type == MOUSEBUTTONDOWN:
				
				#LOGIN/LOGOUT BUTTON:
				if self.login_button.rect.collidepoint(pygame.mouse.get_pos()):
					self.STOP_RUNNING=-1
				if self.logout_button.rect.collidepoint(pygame.mouse.get_pos()):
					self.STOP_RUNNING=-1
				
				#CREDIT BUTTON:
				if self.credit_button.rect.collidepoint(pygame.mouse.get_pos()):
					self.go_credit()
				
				if self.MODE==0:continue #don't allow user to activate any other buttons
				
				
				#LEVEL BUTTONS:
				if self.l1_button.rect.collidepoint(pygame.mouse.get_pos()):self.STOP_RUNNING=1
				if self.l2_button.rect.collidepoint(pygame.mouse.get_pos()):self.STOP_RUNNING=2
				if self.l3_button.rect.collidepoint(pygame.mouse.get_pos()):self.STOP_RUNNING=3
				if self.l4_button.rect.collidepoint(pygame.mouse.get_pos()):self.STOP_RUNNING=4
					
				if self.tuxturn:continue	#don't let player skip, declare ready, or pickup tiles when it's Tux's turn.
				
				#SKIP BUTTON:
				elif self.skip_button.rect.collidepoint(pygame.mouse.get_pos()):
					if self.tuxturn==1:continue# or self.animatedTux.rect[0]<400:continue#MAKE WAIT UNTIL TUX RETURNS
					else:
						
						if self.players[self.player_idx].recording:self.recorder.inform_ready()
						
						self.tuxturn=1
						for spot in self.submissionspots.sprites():#reset
							tile=spot.pop_guest()
							if tile:
								for spot in self.players[1].tray.get_spots():	
									if	tile.saved_center==spot.rect.center:
										spot.take_guest(tile,1)
						
						if self.global_config['soundon']:
							coin=int(random()*len(self.lose_sounds))
							self.lose_sounds[coin].play()
						self.submissionspots.empty()
						self.players[1].actor.queue('go_back')
						return
						
				
				#READY BUTTON:
				elif self.ready_button.rect.collidepoint(pygame.mouse.get_pos()):
					if(self.validator.validate(self.submissionspots.sprites())):#ready for this now!(first player turn)
						
						if self.global_config['soundon']:
							coin=int(random()*len(self.win_sounds))
							self.win_sounds[coin].play()
						
						self.num_replacements=0
						
						self.tuxturn=1
						self.players[1].actor.queue('celebrate')
						self.players[1].actor.queue('go_back')
						self.players[0].actor.queue('get_tougher')
						
						self.board.increment_num_commited()
						self.cannot_move_count=0
						
						for spot in self.submissionspots.sprites():
							for board_spot in self.board.get_spots():
								if board_spot.rect.center==spot.rect.center and spot.guest:
									#validator is gonna have to fill-in AMHEAD fields, then add->board.list_of_heads.add(spot)
									#print spot.guest.str_val,'len=',len(board.get_spots()),len(submissionspots.sprites())
									tile=spot.pop_guest()
									
									#while we've got spot center info set_center of cuckoo_tile:
									rect=board_spot.rect
									xc=rect[0]+rect[2]/2
									yc=rect[1]+rect[3]/2
									self.AnimatedTiles[board_spot.getMN()[0]][board_spot.getMN()[1]].queue('celebrate')
									
									#Validator sets submission_spot and game(this) xfers->board_spot; (i.e. they work together)
									if spot.AMHEAD==1:
										board_spot.AMHEAD=1
										if spot.AMROWEXPR==1:
											board_spot.AMROWEXPR=1
											board_spot.ROWEXPRLENGTH=spot.ROWEXPRLENGTH
											
										if spot.AMCOLEXPR==1:
											board_spot.AMCOLEXPR=1
											board_spot.COLEXPRLENGTH=spot.COLEXPRLENGTH
											
										
									board_spot.take_guest(tile,1)
									board_spot.lock()
									self.submission.remove(spot)
							
						self.move_number+=1
						self.recorder_checkin()
						
						self.draw_tiles()
						self.tuxturn=1
						
						
					else:#put back and play lose sound(not4tux)
                                                if len(self.players[1].actor.maneuver_queue)<2:
                                                    self.players[1].actor.queue('incorrect')
						
						if self.global_config['soundon']:
							coin=int(random()*len(self.lose_sounds))
							self.lose_sounds[coin].play()
						
						for spot in self.submissionspots.sprites():#reset
							tile=spot.pop_guest()
							if tile:
								for spot in self.players[1].tray.get_spots():	
									if	tile.saved_center==spot.rect.center:
										spot.take_guest(tile,1)
						
							
							
						
					self.submissionspots.empty()
					return
				

				
				#check tray for new target:
				for spot in self.players[1].tray.get_spots():
					if not spot.guest:continue
					if spot.rect.collidepoint(pygame.mouse.get_pos()):
						
						if self.global_config['soundon']:
							coin=int(random()*len(self.pickup_sounds))
							self.pickup_sounds[coin].play()
						
						self.target.add(spot.pop_guest())
						
						
				#check the board for new target:
				for spot in self.board.get_spots():
					if spot.rect.collidepoint(pygame.mouse.get_pos()) and spot.islocked():
						
						if self.global_config['soundon']:
							coin=int(random()*len(self.bounce_sounds))
							self.bounce_sounds[coin].play()
						if len(self.players[1].actor.maneuver_queue)<2:
                                                    self.players[1].actor.queue('incorrect')
						
						
				#check the submission for new target:
				#print 'down1:len(submissionspots.sprites())=',len(submissionspots.sprites()),
				for spot in self.submission.get_spots():
					if not spot.occupied():continue
					if spot.rect.collidepoint(pygame.mouse.get_pos()):
						guest=spot.pop_guest()
						self.submissionspots.remove(spot)
						
						if self.global_config['soundon']:
							coin=int(random()*len(self.pickup_sounds))
							self.pickup_sounds[coin].play()
						
						self.target.add(guest)
						
						
				
				
						
			elif event.type == MOUSEBUTTONUP:
				if len(self.target.sprites())>0:
					for spot in self.submission.get_spots():
						if spot.rect.collidepoint(self.target.sprites()[0].rect.center):#welcome to new home
							#print 'up1:len(submissionspots.sprites())=',len(submissionspots.sprites()),
							if spot.occupied():continue#part of submission
							#elif spot.islocked():continue#part of board
							spot.take_guest(self.target.sprites()[0],1)
							
							if self.global_config['soundon']:
								coin=int(random()*len(self.lockin_sounds))
								self.lockin_sounds[coin].play()
							
							self.submissionspots.add(spot)
							self.target.empty()
							
							return
					
					if len(self.target.sprites())==0:continue
					
					for spot in self.board.get_spots():
						if spot.rect.collidepoint(self.target.sprites()[0].rect.center):	
							
							if self.global_config['soundon']:
								coin=int(random()*len(self.bounce_sounds))
								self.bounce_sounds[coin].play()

							if len(self.players[1].actor.maneuver_queue)<2:
                                                            self.players[1].actor.queue('incorrect')

							return
							
					if len(self.target.sprites())>0:#didn't drop? revert!
						#print 'cleanup'
						tile=self.target.sprites()[0]
						self.target.remove(tile)
						for spot in self.players[1].tray.get_spots():	
							if tile.saved_center==spot.rect.center:
								spot.take_guest(tile,1)
								return
		
			#SCREENSHOT
			elif event.type == (KEYUP or KEYDOWN) and event.key==eval(self.global_config['screenshotkey']) and self.env.OS=='lin':
					self.take_screenshot()
					
			elif event.type == (KEYUP or KEYDOWN) and event.key==eval(self.global_config['fullscreenkey']) and self.env.OS=='lin':
				try:pygame.display.toggle_fullscreen()	
				except:pass
			elif event.type == (KEYUP or KEYDOWN) and event.key == K_ESCAPE:
				pygame.quit()
				sys.exit()
			elif event.type == QUIT:self.STOP_RUNNING=-1#unused?
			
	
	def take_screenshot(self):
		display_surface=pygame.display.get_surface()
		tstamp=self.mktstamp()
		oufname="tuxmathscrabble_%s.bmp"%(tstamp)
		if self.env.OS=='lin':
                    oufname=os.path.join(os.environ['HOME'],oufname)
		pygame.image.save(display_surface,oufname)
		self.screenshot_counter+=1
		#print 'SCREENSHOT TAKEN:',oufname

	def mktstamp(self):
		#tstamp which increases monotonically with time
		t=time.localtime()
		YYYY="%d"%t[0]
		MM="%02d"%t[1]
		DD="%02d"%t[2]
		hh="%02d"%t[3]
		mm="%02d"%t[4]
		ss="%02d"%t[5]
		tstamp="%s%s%s%s%s%s"%(YYYY,MM,DD,hh,mm,ss)
		return tstamp
	
	def queue_thinking_maneuver(self):
		self.players[self.player_idx].actor.queue('thinking')
		self.animation_in_progress=1
		while self.animation_in_progress:
			self.update()
			
	def tux_wait_maneuver(self):
			#coin=int(random()*len(self.user_config['wait_maneuvers']))
			#self.animator.start_maneuver('Tux',self.user_config['wait_maneuvers'][coin])
			#self.update()
			pass
			
	def update(self):
			
			
			if not self.tuxturn:who_is_waiting=0
			else:who_is_waiting=1
			self.players[who_is_waiting].wait_counter+=1
			if self.players[who_is_waiting].wait_counter>=self.global_config['wait_counter']:
				self.players[who_is_waiting].wait_counter=0
				self.players[who_is_waiting].actor.queue('waiting')
				
			self.handle_mouse()
				
			self.screen.blit(self.background,(0,0))
			
			#
			b=self.l1_button
			if self.LEVEL==2:	b=self.l2_button
			elif self.LEVEL==3:	b=self.l3_button
			elif self.LEVEL==4:	b=self.l4_button
			self.screen.blit(self.level_indicator,(10,b.rect[1]+b.rect[3]/3))
			
			#Scoreboard:
			self.tuxscorestring="Tux:%03d"%self.players[0].score
			self.screen.blit(self.pyfont.render(self.tuxscorestring,1,self.global_config['fgcolor'],self.global_config['bgcolor']),(20,10))
			self.playerscorestring="You:%03d"%self.players[1].score
			self.screen.blit(self.pyfont.render(self.playerscorestring,1,self.global_config['fgcolor'],self.global_config['bgcolor']),(20,30))
			
			#Who's Turn-Indicator:
			XPOS=self.global_config['WIN_W']/2+(self.players[0].tray.N*self.global_config['tilesize'][0])/2
			if self.LEVEL>=3:
				if self.tuxturn==1:
					tuxturnstring= '<-Tux'
					self.screen.blit(self.pyfont.render(tuxturnstring,1,self.global_config['fgcolor'],self.global_config['bgcolor']),(XPOS,20))
				else:
					yourturnstring='<-You'
					self.screen.blit(self.pyfont.render(yourturnstring,1,self.global_config['fgcolor'],self.global_config['bgcolor']),(XPOS,self.global_config['WIN_H']-65))
			else:
				if self.tuxturn==1:
					tuxturnstring= '<-Tux'
					self.screen.blit(self.pyfont.render(tuxturnstring,1,self.global_config['fgcolor'],self.global_config['bgcolor']),(XPOS,20))
				else:
					yourturnstring='<-You'
					self.screen.blit(self.pyfont.render(yourturnstring,1,self.global_config['fgcolor'],self.global_config['bgcolor']),(XPOS,self.global_config['WIN_H']-65))
			
			
			self.buttons.draw(self.screen)
			if self.MODE==0:self.loginbuttons.draw(self.screen)
			
			self.boardspots.draw(self.screen)
			self.submissionspots.draw(self.screen)
			
			
			for idx in range(2):
				#pygame.sprite.RenderPlain(self.players[idx].tray.get_spots()).draw(self.screen)
				self.tray_spots[idx].draw(self.screen)
			
			if len(self.target.sprites())>0:
				self.target.sprites()[0].update()
				self.target.draw(self.screen)
			
			t=time.time()
			self.animation_in_progress=0
			
			#ANIMATED TILE UPDATES:
			for midx in range(len(self.AnimatedTiles)):
				for nidx in range(len(self.AnimatedTiles[0])):
					if self.AnimatedTiles[midx][nidx].xdest or len(self.AnimatedTiles[midx][nidx].maneuver_queue):
						self.AnimatedTiles[midx][nidx].render(t,self.screen,None,None)
						self.animation_in_progress=1

			#ACTOR UPDATES:
			for plyr in self.players:
				plyr.actor.render(t,self.screen,None,None)
				if plyr.actor.xdest:self.animation_in_progress=1
			
			
			#HIGH_SCORE_LIST:
			if self.MODE==0 and self.global_config['high_score_list']:self.blit_high_scores()
					
			#self.fwoverlay.tick()
			
			pygame.display.flip()
			
			if self.global_config['record_movie']:
				oufname="%05d.bmp"%(self.movie_idx)
				pygame.image.save(pygame.display.get_surface(),oufname)
				self.movie_idx+=1


	def go_credit(self):			
		credit_image=pygame.image.load(os.path.join(self.env.sitepkgdir,self.global_config['appdir'],'Images','credit_splash.gif'))
		credit_image=pygame.transform.scale(credit_image,(self.screen.get_width(),self.screen.get_height()))
		self.screen.fill((0,0,0))
		self.screen.blit(credit_image,(0,0))
		pygame.display.flip()
		while 1:
			breakout=0
			for event in pygame.event.get():
				if event.type == KEYDOWN:breakout=1
				elif event.type == MOUSEBUTTONDOWN:breakout=1
			if breakout:break
		
		if self.fwoverlay:
			self.fwoverlay.reset()
			self.fwoverlay.make_volley(self.global_config['fw_volley_size'])
			
		return 1	


	def recorder_checkin(self):
		#update all variables specified as options in user_config (i.e. "new")
		t=time.time()
		self.dt_last=t-self.t_last
		self.t_session+=self.dt_last
		self.t_last=t

		#self.last_points set from tms_solver and/or tms_validator
		#self.num_replacements set above
		
		try:
			self.num_new_tiles=len(self.solver_submission)
		except:
			self.num_new_tiles=0
		
		self.move_number=self.move_number
		
		self.IDNstate	=self.global_config['IDNstate']
		self.IDNdistrict=self.global_config['IDNdistrict']
		self.IDNschool	=self.global_config['IDNschool']
		self.IDNgrade	=self.user_config['IDNgrade']
		self.IDNsession	=self.user_config['IDNsession']
		self.IDNstudent	=self.user_config['IDNstudent']
		self.IDNteacher	=self.user_config['IDNteacher']
		
		
		if self.players[self.player_idx].recording:self.recorder.inform_ready()


	def blit_high_scores(self):		
			H0=60
			dH0=60
			#self.pyfont.set_bold(1)
			font_surface=self.congrats_font.render(self.global_config['high_score_string'],1,self.global_config['high_score_color'],self.global_config['bgcolor'])
			font_surface.set_colorkey(self.global_config['bgcolor'])
			self.screen.blit(font_surface,(self.W/2-font_surface.get_width()/2,H0))
			H0 += dH0
			for idx in range(0,len(self.high_score_list)):
				hs_string="%s: %d"%(self.high_score_list[idx][0],self.high_score_list[idx][1])
				font_surface=self.congrats_font.render(hs_string,1,self.global_config['high_score_color'],self.global_config['bgcolor'])
				font_surface.set_colorkey(self.global_config['bgcolor'])
				self.screen.blit(font_surface,(self.W/2-font_surface.get_width()/2,H0))
				H0 += dH0
				if H0>self.W-2*font_surface.get_height():break
			#self.pyfont.set_bold(0)
